/*
 * Test DMA transfers.
 *
 * Copyright (C) 2018, 2020, 2021, 2023 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 */

#include <l4/devices/dma.h>
#include <l4/devices/cpm-jz4730.h>
#include <l4/devices/dma-jz4730.h>
#include <l4/devices/memory.h>

#include <l4/re/c/util/cap_alloc.h>
#include <l4/re/c/dataspace.h>
#include <l4/re/c/rm.h>

#include <l4/sys/err.h>
#include <l4/sys/factory.h>
#include <l4/sys/icu.h>
#include <l4/sys/irq.h>
#include <l4/sys/rcv_endpoint.h>

#include <stdio.h>
#include <string.h>
#include <unistd.h>



/* Device and resource discovery. */

static long item_in_range(long start, long end, long index)
{
  if (start < end)
    return start + index;
  else
    return start - index;
}



int main(void)
{
  long err;
  void *cpm;
  void *dmac, *dma0;

  /* Allocate memory to test transfers. */

  l4_cap_idx_t ds0_mem, ds1_mem;
  l4_size_t ds0_size = L4_PAGESIZE, ds1_size = L4_PAGESIZE;
  l4_addr_t ds0_addr, ds1_addr;
  l4re_dma_space_dma_addr_t ds0_paddr, ds1_paddr;

  err = get_dma_region(ds0_size, 8, &ds0_addr, &ds0_paddr, &ds0_mem);

  if (err)
  {
    printf("Could not allocate region #0.\n");
    return 1;
  }

  err = get_dma_region(ds1_size, 8, &ds1_addr, &ds1_paddr, &ds1_mem);

  if (err)
  {
    printf("Could not allocate region #1.\n");
    return 1;
  }

  /* Fill the allocated memory. */

  memset((void *) ds0_addr, 0, ds0_size);
  memset((void *) ds1_addr, 0, ds1_size);

  sprintf((char *) ds0_addr, "The quick brown fox jumped over the lazy dog.\n");

  /* Interrupts. */

  l4_uint32_t dma_irq_start = 0, dma_irq_end = 0;
  l4_cap_idx_t icucap, irq0cap;

  /* Obtain resource details describing the interrupt for DMA channel 0. */

  printf("Access IRQ...\n");

  if (get_irq("jz4730-dma", &dma_irq_start, &dma_irq_end) < 0)
    return 1;

  printf("IRQ range at %d...%d.\n", dma_irq_start, dma_irq_end);

  /* Obtain capabilities for the interrupt controller and an interrupt. */

  irq0cap = l4re_util_cap_alloc();
  icucap = l4re_env_get_cap("icu");

  if (l4_is_invalid_cap(icucap))
  { 
    printf("No 'icu' capability available in the virtual bus.\n");
    return 1;
  }

  if (l4_is_invalid_cap(irq0cap))
  {
    printf("Capabilities not available for interrupts.\n");
    return 1;
  }

  /* Create interrupt objects. */

  err = l4_error(l4_factory_create_irq(l4re_global_env->factory, irq0cap));

  if (err)
  {
    printf("Could not create IRQ object: %lx\n", err);
    return 1;
  }

  /* Bind interrupt objects to IRQ numbers. */

  err = l4_error(l4_icu_bind(icucap,
                             item_in_range(dma_irq_start, dma_irq_end, 0),
                             irq0cap));

  if (err)
  {
    printf("Could not bind IRQ to the ICU: %ld\n", err);
    return 1;
  }

  /* Attach ourselves to the interrupt handler. */

  err = l4_error(l4_rcv_ep_bind_thread(irq0cap, l4re_env()->main_thread, 0));

  if (err)
  {
    printf("Could not attach to IRQs: %ld\n", err);
    return 1;
  }

  /* Peripheral memory. */

  l4_addr_t cpm_base = 0, cpm_base_end = 0;
  l4_addr_t dma_base = 0, dma_base_end = 0;

  /* Obtain resource details describing I/O memory. */

  printf("Access CPM...\n");

  if (get_memory("jz4730-cpm", &cpm_base, &cpm_base_end) < 0)
    return 1;

  printf("CPM at 0x%lx...0x%lx.\n", cpm_base, cpm_base_end);

  printf("Access DMA...\n");

  if (get_memory("jz4730-dma", &dma_base, &dma_base_end) < 0)
    return 1;

  printf("DMA at 0x%lx...0x%lx.\n", dma_base, dma_base_end);

  /* Obtain CPM and DMA references. */

  cpm = jz4730_cpm_init(cpm_base);
  dmac = jz4730_dma_init(dma_base, dma_base_end, cpm);
  dma0 = jz4730_dma_get_channel(dmac, 0, irq0cap);

  /* Enable DMA. */

  printf("Enable DMA...\n");

  jz4730_dma_enable(dmac);

  /* Transfer data between the allocated memory regions. */

  printf("Transfer from %llx to %llx...\n", ds0_paddr, ds1_paddr);

  unsigned int count = L4_PAGESIZE / 4;

  unsigned int to_transfer = jz4730_dma_transfer(dma0, (uint32_t) ds0_paddr,
                                                 (uint32_t) ds1_paddr,
                                                 count,
                                                 1, 1,
                                                 4, 4,
                                                 4,
                                                 Dma_request_auto);

  unsigned int transferred = to_transfer ? count - jz4730_dma_wait(dma0) : 0;

  printf("Transferred: %d\n", transferred);
  printf("Source: %s\n", (char *) ds0_addr);
  printf("Destination: %s\n", (char *) ds1_addr);

  /* Detach from the interrupt. */

  err = l4_error(l4_irq_detach(irq0cap));

  if (err)
    printf("Error detaching from IRQ objects: %ld\n", err);

  return 0;
}
